
from __future__ import annotations
import math
from typing import Dict, Any, Optional
import tensorflow as tf
from returnn.tf.layers.basic import LinearLayer
from .module import Module
from ..parameter import Parameter
from ...tensor import Tensor
from .. import init


class Embedding(Module):

  def __init__(self, num_embeddings: int, embedding_dim: int, padding_idx: Optional[int] = None,
               max_norm: Optional[float] = None, norm_type: float = 2., scale_grad_by_freq: bool = False,
               sparse: bool = False, _weight: Optional[Tensor] = None) -> None:
    super(Embedding, self).__init__()
    self.num_embeddings = num_embeddings
    self.embedding_dim = embedding_dim
    if padding_idx is not None:
      if padding_idx > 0:
        assert padding_idx < self.num_embeddings, 'Padding_idx must be within num_embeddings'
      elif padding_idx < 0:
        assert padding_idx >= -self.num_embeddings, 'Padding_idx must be within num_embeddings'
        padding_idx = self.num_embeddings + padding_idx
    self.padding_idx = padding_idx
    self.max_norm = max_norm
    self.norm_type = norm_type
    self.scale_grad_by_freq = scale_grad_by_freq
    if _weight is None:
      self.weight = Parameter(Tensor(num_embeddings, embedding_dim))
      self.reset_parameters()
    else:
      assert list(_weight.shape) == [num_embeddings, embedding_dim], (
        'Shape of weight does not match num_embeddings and embedding_dim')
      self.weight = Parameter(_weight)
    self.sparse = sparse

  def reset_parameters(self) -> None:
    init.normal_(self.weight)
    if self.padding_idx is not None:
      self.weight[self.padding_idx].fill_(0)

  @classmethod
  def from_pretrained(cls, embeddings, freeze=True, padding_idx=None,
                      max_norm=None, norm_type=2., scale_grad_by_freq=False,
                      sparse=False):
    assert embeddings.dim() == 2, 'Embeddings parameter is expected to be 2-dimensional'
    rows, cols = embeddings.shape
    embedding = cls(
      num_embeddings=rows,
      embedding_dim=cols,
      _weight=embeddings,
      padding_idx=padding_idx,
      max_norm=max_norm,
      norm_type=norm_type,
      scale_grad_by_freq=scale_grad_by_freq,
      sparse=sparse)
    embedding.weight.requires_grad = not freeze
    return embedding

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    return {
      "class": "linear", "from": self._get_input_layer_name(input),
      "activation": None, "with_bias": False,
      "n_out": self.embedding_dim}

  def check_returnn_layer(self, layer: LinearLayer):
    return layer.input_data.sparse and layer.input_data.dim == self.num_embeddings

  def import_params_torch_to_returnn(self, *, layer: LinearLayer, torch_module: Embedding):
    session = tf.compat.v1.get_default_session()
    values = torch_module.weight.detach().numpy()
    layer.params["W"].load(values, session=session)


__all__ = [
  key for (key, value) in sorted(globals().items())
  if not key.startswith("_")
  and getattr(value, "__module__", "") == __name__]




from __future__ import annotations
import math
from typing import Dict, Any, Tuple, List
import tensorflow as tf
from returnn.tf.layers.basic import LayerBase, LinearLayer
from .module import Module
from ..parameter import Parameter
from ...tensor import Tensor
from .. import init
from ....naming import Naming, TensorEntry


class Identity(Module):
  def __init__(self, *args, **kwargs):
    super(Identity, self).__init__()

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    return {"class": "copy", "from": self._get_input_layer_name(input)}


class Linear(Module):
  in_features: int
  out_features: int
  weight: Tensor

  def __init__(self, in_features: int, out_features: int, bias: bool = True) -> None:
    super(Linear, self).__init__()
    self.in_features = in_features
    self.out_features = out_features
    self.weight = Parameter(Tensor(out_features, in_features))
    if bias:
      self.bias = Parameter(Tensor(out_features))
    else:
      self.register_parameter('bias', None)
    self.reset_parameters()

  def reset_parameters(self) -> None:
    init.kaiming_uniform_(self.weight, a=math.sqrt(5))
    if self.bias is not None:
      fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
      bound = 1 / math.sqrt(fan_in)
      init.uniform_(self.bias, -bound, bound)

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    return {
      "class": "linear", "from": self._get_input_layer_name(input),
      "n_out": self.out_features,
      "with_bias": self.bias is not None,
      "activation": None}

  def check_returnn_layer(self, layer: LinearLayer):
    assert layer.input_data.dim == self.in_features

  def import_params_torch_to_returnn(self, *, layer: LinearLayer, torch_module: Linear):
    session = tf.compat.v1.get_default_session()
    values = torch_module.weight.detach().numpy()
    values = values.transpose()
    layer.params["W"].load(values, session=session)
    if self.bias is not None:
      layer.params["b"].load(torch_module.bias.detach().numpy(), session=session)



class Matmul(Module):
  """
  Maps to RETURNN DotLayer
  """
  is_original_torch_module = False

  def _get_output_shape_from_returnn(self, inputs_flat: List[Tensor], layer: LayerBase
                                     ) -> Tuple[Tuple[int, ...], Dict[int, int]]:
    assert len(inputs_flat) == 2
    input1, input2 = inputs_flat
    assert all(isinstance(input_, Tensor) for input_ in [input1, input2])
    assert all(len(input_.shape) >= 2 for input_ in [input1, input2]), "not implemented otherwise"

    max_len = max(len(input1.shape), len(input2.shape))
    shape1 = [None] * (max_len - len(input1.shape)) + list(input1.shape)
    shape2 = [None] * (max_len - len(input2.shape)) + list(input2.shape)
    shape = []
    for ax in range(max_len - 2):
      if shape1[ax] == shape2[ax]:
        shape.append(shape1[ax])
      elif isinstance(shape1[ax], int) and shape1[ax] > 1:
        assert shape2[ax] in [1, None], f"dimensions are not broadcastable: {input1.shape} vs. {input2.shape}"
        shape.append(shape1[ax])
      elif isinstance(shape2[ax], int) and shape2[ax] > 1:
        assert shape1[ax] in [1, None], f"dimensions are not broadcastable: {input1.shape} vs. {input2.shape}"
        shape.append(shape2[ax])

    shape += [shape1[-2], shape2[-1]]
    shape = tuple(shape)
    returnn_axis_from_torch_axis = {i: i for i in range(len(shape))}
    return shape, returnn_axis_from_torch_axis

  def create_returnn_layer_dict(self, *inputs: Tensor, **kwargs) -> Dict[str, Any]:
    sources = [self._get_input_layer_name(source) for source in inputs]
    assert len(sources) == 2

    assert len(inputs[0].shape) >= 2, "not implemented otherwise"
    assert len(inputs[1].shape) >= 2, "not implemented otherwise"
    red1 = self._get_input_axis_to_returnn(inputs[0], -1)
    red2 = self._get_input_axis_to_returnn(inputs[1], -2)
    var1 = [self._get_input_axis_to_returnn(inputs[0], -2)]
    var2 = [self._get_input_axis_to_returnn(inputs[1], -1)]

    max_len = max(len(inputs[0].shape), len(inputs[1].shape))
    shape1 = [None] * (max_len - len(inputs[0].shape)) + list(inputs[0].shape)
    shape2 = [None] * (max_len - len(inputs[1].shape)) + list(inputs[1].shape)
    for ax in range(-3, -max_len - 1, -1):
      if isinstance(shape1[ax], int) and shape1[ax] > 1:
        if shape2[ax] in [1, None]:
          var1.insert(0, self._get_input_axis_to_returnn(inputs[0], ax))
        else:
          assert shape1[ax] == shape2[ax], f"dimensions are not broadcastable: {inputs[0].shape} vs. {inputs[1].shape}"
      if isinstance(shape2[ax], int) and shape2[ax] > 1:
        if shape1[ax] in [1, None]:
          var2.insert(0, self._get_input_axis_to_returnn(inputs[1], ax))
        else:
          assert shape1[ax] == shape2[ax], f"dimensions are not broadcastable: {inputs[0].shape} vs. {inputs[1].shape}"

    return {"class": "dot", "red1": red1, "red2": red2, "var1": var1, "var2": var2, "from": sources}


__all__ = [
  key for (key, value) in sorted(globals().items())
  if not key.startswith("_")
  and getattr(value, "__module__", "") == __name__]




from __future__ import annotations
from typing import Optional, List, Dict, Tuple, Any
from collections import OrderedDict
import itertools
from . import call as _call
from . import module as _module
from . import tensor as _tensor
from . import returnn_ctx as _returnn_ctx
from . import naming as _naming


class RegisteredName:
  childs_by_name: OrderedDict[str, RegisteredName]
  parent: Optional[RegisteredName]
  name: Optional[str]  # if parent
  level: int = 0
  modules: List[_module.ModuleEntry]  # can be multiple merged together
  calls: List[_call.CallEntry]  # can be multiple merged together. can be empty if this is some input
  tensor: Optional[_tensor.TensorEntry] = None  # output from the call
  returnn_ctx: Optional[_returnn_ctx.ReturnnContext] = None
  is_reserved: bool  # e.g. input "data" or output "output"
  is_subnet: bool
  is_alternative_subnet_output: bool = False
  ReservedInputName = "data"
  ReservedOutputName = "output"
  ReservedNames = {ReservedInputName, ReservedOutputName}
  _inputs: List[RegisteredName]

  def __init__(self, *,
               wrap_to_returnn_enabled: Optional[bool] = None,
               returnn_train_flag: Optional[bool] = None,
               parent: Optional[RegisteredName] = None, name: Optional[str] = None,
               call: Optional[_call.CallEntry] = None,
               tensor: Optional[_tensor.TensorEntry] = None,
               is_reserved: bool = False, is_subnet: bool):
    self.childs_by_name = OrderedDict()
    self._inputs = []
    self.parent = parent
    if parent:
      assert name
      assert wrap_to_returnn_enabled is None
      assert returnn_train_flag is None
      wrap_to_returnn_enabled = parent.wrap_to_returnn_enabled
      returnn_train_flag = parent.returnn_train_flag
    else:
      assert not name
      assert wrap_to_returnn_enabled is not None
      assert returnn_train_flag is not None
    self.name = name
    self.wrap_to_returnn_enabled = wrap_to_returnn_enabled
    self.returnn_train_flag = returnn_train_flag
    self.is_reserved = is_reserved
    if not is_reserved:
      assert name not in self.ReservedNames
      # If reserved, we also allow other names...
    self.is_subnet = is_subnet
    self.modules = []
    self.calls = []
    if call:
      self.assign_call(call)
    if parent:
      self.level = parent.level + 1
    if tensor:
      self.assign_tensor(tensor)
    if self.wrap_to_returnn_enabled:
      if not parent:  # with parent, returnn_ctx will be created once needed
        self.maybe_create_returnn_ctx()

  def __repr__(self):
    return f"<{self.__class__.__name__} {self.get_absolute_name()!r} {self._repr_content()}>"

  def _repr_content(self) -> str:
    if len(self.modules) == 0:
      mod = None
    elif len(self.modules) == 1:
      mod = self.modules[0]
    else:
      mod = self.modules
    if len(self.calls) == 0:
      res = None
    elif len(self.calls) == 1:
      res = self.calls[0].outputs_flat
      if res is None:
        res = "..."
      elif len(res) == 0:
        res = f"<{self.calls[0]} without outputs>"
      elif len(res) == 1:
        res = res[0]
    else:
      res = f"<multiple calls {self.calls}>"
    return f"{mod} -> {res}"

  def get_absolute_name(self):
    names = []
    name_ = self
    while name_.parent:
      names.insert(0, name_.name)
      name_ = name_.parent
    return "/".join(names) if names else ""

  def get_parents_hierarchy(self) -> List[RegisteredName]:
    res = []
    x = self
    while x.parent:
      if x.name.startswith("."):  # stop at hidden
        break
      res.append(x.parent)
      x = x.parent
    return res

  def assign_tensor(self, tensor: _tensor.TensorEntry):
    if self.tensor:
      self.tensor.names.remove(self)
    self.tensor = tensor
    tensor.names.append(self)

  def assign_call(self, call: _call.CallEntry):
    if call in self.calls:
      return
    self.assign_module(call.module)
    if self.is_subnet:
      assert call.module.module.has_torch_forward()
    else:
      assert not self.calls  # cannot have multiple calls assigned
      assert not call.module.module.has_torch_forward()
    assert not call.namespace
    call.namespace = self
    self.calls.append(call)

  def maybe_create_returnn_ctx(self):
    """
    Makes sure that returnn_ctx is created.
    """
    assert self.is_subnetwork()
    assert self.wrap_to_returnn_enabled
    if self.returnn_ctx:
      return
    if self.parent:
      if not self.parent.returnn_ctx:
        self.parent.maybe_create_returnn_ctx()
      assert self.parent.returnn_ctx
    self.returnn_ctx = _returnn_ctx.ReturnnContext(
      parent=self.parent.returnn_ctx if self.parent else None,
      name=self.name, returnn_train_flag=self.returnn_train_flag)

  def assign_module(self, module: _module.ModuleEntry):
    if module in self.modules:
      return
    if self.is_subnet:
      assert module.module.has_torch_forward()
    else:
      assert not self.modules  # cannot have multiple modules assigned
      assert not module.module.has_torch_forward()
    self.modules.append(module)
    module.names.append(self)
    if self.is_reserved:
      assert not module.module.has_torch_forward()
    if self.wrap_to_returnn_enabled:
      if module.module.has_torch_forward() and not self.returnn_ctx:
        # Need our own returnn ctx / subnet.
        assert self.parent
        self.maybe_create_returnn_ctx()

  def is_subnetwork(self) -> bool:
    """
    If True, this would be wrapped as a subnetwork in RETURNN.
    If False, this directly maps to a layer in RETURNN.
    """
    return self.is_subnet

  def _get_unique_name(self, suggested_name: str) -> str:
    if suggested_name not in self.childs_by_name and suggested_name not in self.ReservedNames:
      return suggested_name
    for i in itertools.count(1):
      suggested_name_ = f"{suggested_name}_{i}"
      if suggested_name_ not in self.childs_by_name and suggested_name_ not in self.ReservedNames:
        return suggested_name_

  def register_sub_net(self, *, suggested_name: str) -> RegisteredName:
    assert self.is_subnetwork()
    name = self._get_unique_name(suggested_name)
    child = RegisteredName(parent=self, name=name, is_subnet=True)
    self.childs_by_name[name] = child
    return child

  def register_sub_call(self, call: _call.CallEntry) -> RegisteredName:
    assert self.is_subnetwork()
    name = self._get_unique_name(call.module.get_canonical_name(parent_namespace=self))
    child = RegisteredName(parent=self, name=name, is_subnet=call.module.module.has_torch_forward())
    self.childs_by_name[name] = child
    child.assign_call(call)
    return child

  def register_input(self, tensor: _tensor.TensorEntry) -> RegisteredName:
    assert self.is_subnetwork()
    name = self.ReservedInputName
    assert tensor not in self._inputs
    idx = len(self._inputs)
    if idx != 0:
      name += f":{idx}"  # should be consistent with RETURNN SubnetworkLayer concat_sources=False input naming logic
    assert name not in self.childs_by_name
    name_ = RegisteredName(parent=self, name=name, tensor=tensor, is_reserved=True, is_subnet=False)
    self.childs_by_name[name] = name_
    self._inputs.append(name_)
    if self.wrap_to_returnn_enabled:
      self.returnn_ctx.define_input(tensor, data_key=str(idx) if idx else None)
    return name_

  def register_returnn_subnet_output(self, tensor: _tensor.TensorEntry) -> RegisteredName:
    assert self.is_subnetwork()
    from pytorch_to_returnn.torch.nn import Copy
    naming = _naming.Naming.get_instance()
    assert naming.wrap_to_returnn_enabled
    _, reg_name = self.name_for_tensor(tensor)
    potential_calls = set(tensor.output_from_calls).intersection(set(reg_name.calls))
    assert len(potential_calls) == 1, f"{tensor.output_from_calls} vs {reg_name.calls}"
    call = list(potential_calls)[0]
    self.assign_tensor(tensor)  # for this subnet
    name = self.ReservedOutputName
    assert name not in self.childs_by_name
    child = RegisteredName(parent=self, name=name, is_reserved=True, is_subnet=False)
    self.childs_by_name[name] = child
    copy_mod = Copy()
    copy_call = _call.CallEntry(module=naming.modules[copy_mod])
    copy_call.parent_call = call
    call.child_calls.append(copy_call)
    copy_call.inputs_args = (tensor,)
    copy_call.inputs_flat = [tensor]
    copy_call.inputs_kwargs = {}
    child.assign_call(copy_call)
    naming.module_call_stack.append(copy_call)
    try:
      copy_call.apply_call()
    finally:
      assert copy_call is naming.module_call_stack[-1]
      naming.module_call_stack.pop(-1)
    self.returnn_ctx.define_output(copy_call.returnn_layer)
    tensor.returnn_data = copy_call.returnn_layer.output
    return child

  def name_in_ctx(self, possible_sub_names: List[RegisteredName],
                  _src_tensor: Optional[_tensor.TensorEntry] = None) -> Tuple[str, RegisteredName]:
    assert self.is_subnetwork()
    # First try to find in our direct namespace.
    # Only if that fails, search in potential sub namespaces.
    parent_idx = 0
    while True:
      parent_indices = []
      for name_ in possible_sub_names:
        parent_hierarchy = name_.get_parents_hierarchy()
        if self not in parent_hierarchy:
          continue
        if parent_hierarchy[parent_idx] is self:
          if parent_idx == 0:
            assert self.childs_by_name[name_.name] is name_
            return name_.name, name_
          name_.is_alternative_subnet_output = True
          return "/".join([parent.name for parent in reversed(parent_hierarchy[:parent_idx])] + [name_.name]), name_
        parent_indices.append(parent_hierarchy.index(self))
      if not parent_indices:
        # If you get here, check the logic in Module.__call__, Naming.push_module_call.
        raise KeyError(f"namespace {self!r}: {_src_tensor or possible_sub_names!r} not found")
      assert parent_idx == 0  # should not get here twice
      parent_idx = min(parent_indices)

  def name_for_tensor(self, tensor: _tensor.TensorEntry) -> Tuple[str, RegisteredName]:
    assert self.is_subnetwork()
    assert tensor.names
    return self.name_in_ctx(possible_sub_names=tensor.names, _src_tensor=tensor)

  def find_name_for_module(self, module: _module.ModuleEntry) -> Optional[str]:
    assert self.is_subnetwork()
    for name, child in self.childs_by_name.items():
      if module in child.modules:
        return name
    return None

  def dump(self, prefix=""):
    for name, child in self.childs_by_name.items():
      if name.startswith("."):
        print(f"{prefix}{name}: (hidden, {'non-empty' if child.childs_by_name else 'empty'})")
        continue
      print(f"{prefix}{name}: {child._repr_content()}")
      child.dump(prefix=f"{prefix}  ")

  def dump_as_returnn_layer_dict(self) -> Dict[str, Any]:
    if not self.is_subnet:
      assert len(self.calls) == 1
      call = self.calls[0]
      assert call.returnn_layer_dict
      return call.returnn_layer_dict
    # Subnetwork
    inputs = []
    for input_child in self._inputs:
      input_tensor = input_child.tensor
      assert input_tensor
      assert self.parent
      parent_namespace = self.parent
      input_layer_name, _ = parent_namespace.name_for_tensor(input_tensor)
      inputs.append(input_layer_name)
    subnet_dict = self.dump_as_returnn_net_dict()
    if len(inputs) <= 1:
      return {"class": "subnetwork", "from": inputs[0] if inputs else [], "subnetwork": subnet_dict}
    return {
      "class": "subnetwork", "from": inputs, "concat_sources": False, "subnetwork": subnet_dict}

  def dump_as_returnn_net_dict(self) -> Dict[str, Dict[str, Any]]:
    net_dict = {}
    for name, child in self.childs_by_name.items():
      if not child.calls:
        continue  # e.g. input "data"
      net_dict[name] = child.dump_as_returnn_layer_dict()
      if child.is_alternative_subnet_output:
        net_dict[name]["is_output_layer"] = True
    return net_dict





import tensorflow as tf
import torch
import contextlib
import numpy
import types
import tempfile
from pytorch_to_returnn.pprint import pprint
from typing import Callable, Optional, Dict, Any, Tuple, List
from returnn.tf.util.data import Data
from pytorch_to_returnn import torch as torch_returnn
from pytorch_to_returnn.import_wrapper import wrapped_import_torch_traced, wrapped_import_torch_returnn
from pytorch_to_returnn.import_wrapper.torch_wrappers.tensor import WrappedTorchTensor
from pytorch_to_returnn.naming import Naming


ModelFuncType = Callable[[Optional[Callable[[str], types.ModuleType]], torch.Tensor], torch.Tensor]


@contextlib.contextmanager
def make_scope():
  """
  :rtype: tf.compat.v1.Session
  """
  with tf.Graph().as_default() as graph:
    with tf.compat.v1.Session(graph=graph) as session:
      yield session


class Converter:
  """
  Example code for model func::

    def model_func(wrapped_import, inputs):

        if typing.TYPE_CHECKING or not wrapped_import:
            import torch
            from parallel_wavegan import models as pwg_models
            from parallel_wavegan import layers as pwg_layers

        else:
            torch = wrapped_import("torch")
            wrapped_import("parallel_wavegan")
            pwg_models = wrapped_import("parallel_wavegan.models")
            pwg_layers = wrapped_import("parallel_wavegan.layers")

        # Initialize PWG
        pwg_config = yaml.load(open(args.pwg_config), Loader=yaml.Loader)
        pyt_device = torch.device("cpu")
        generator = pwg_models.MelGANGenerator(**pwg_config['generator_params'])
        generator.load_state_dict(
            torch.load(args.pwg_checkpoint, map_location="cpu")["model"]["generator"])
        generator.remove_weight_norm()
        pwg_model = generator.eval().to(pyt_device)
        assert pwg_config["generator_params"].get("aux_context_window", 0) == 0  # not implemented otherwise
        pwg_pqmf = pwg_layers.PQMF(pwg_config["generator_params"]["out_channels"]).to(pyt_device)

        with torch.no_grad():
            return pwg_pqmf.synthesis(pwg_model(inputs))
  """

  def __init__(self,
               model_func: ModelFuncType, *,
               inputs: numpy.ndarray,
               inputs_data_kwargs: Optional[Dict[str, Any]] = None,
               returnn_dummy_input_shape: Optional[Tuple[int]] = None,
               returnn_input_seq_lens: Optional[Dict[int, List[int]]] = None,
               use_non_wrapped_reference: bool = True,
               verify_with_torch: bool = True,
               verify_individual_model_io: bool = True,
               import_torch_params: bool = True,
               export_tf_checkpoint_save_path: Optional[str] = None,
               verify_returnn_standalone_model: bool = True,
               train: bool = False,
               ):
    """
    :param model_func:
      Gets an argument wrapped_import(str) -> module, or None. If None, should import as is.
      It also gets the inputs, converted to the right PyTorch `Tensor` object (either original or wrapped).
    :param inputs: example inputs
    """
    self._model_func = model_func
    self._inputs_np = inputs
    inputs_data_kwargs = {} if inputs_data_kwargs is None else inputs_data_kwargs.copy()
    if "feature_dim_axis" not in inputs_data_kwargs and not inputs_data_kwargs.get("sparse", False):
      assert len(inputs.shape) >= 2  # (batch,feature|channel,...)
      if inputs_data_kwargs.get("batch_dim_axis", 0) == 0:
        inputs_data_kwargs["feature_dim_axis"] = 1  # Torch uses batch-feature-major by default
    if "shape" not in inputs_data_kwargs:
      # Assume feature is static, all other dynamic.
      # Assume batch-major.
      inputs_data_kwargs["shape"] = [
        inputs.shape[i] if i == inputs_data_kwargs["feature_dim_axis"] else None
        for i in range(1, len(inputs.shape))]
    self._returnn_in_data_dict = inputs_data_kwargs
    self._returnn_dummy_input_shape = returnn_dummy_input_shape
    self._returnn_input_seq_lens = returnn_input_seq_lens
    self.use_non_wrapped_reference = use_non_wrapped_reference
    self.verify_with_torch = verify_with_torch
    self.verify_individual_model_io = verify_individual_model_io
    self.import_torch_params = import_torch_params
    self.export_tf_checkpoint_save_path = export_tf_checkpoint_save_path
    self.train = train
    self._tf_checkpoint_save_path = None  # type: Optional[str]
    self.verify_returnn_standalone_model = verify_returnn_standalone_model
    self._out_ref_np = None  # type: Optional[numpy.ndarray]
    self._torch_namespace = None  # type: Optional[Naming]
    self._out_returnn_np = None  # type: Optional[numpy.ndarray]
    self._returnn_net_dict = None  # type: Optional[Dict[str, Dict[str, Any]]]

  def run(self):
    if self.use_non_wrapped_reference:
      self._run_reference()
    if self.verify_with_torch or self.verify_individual_model_io or self.import_torch_params:
      self._run_traced_orig_torch()
    self._run_torch_returnn_drop_in()
    if self.verify_returnn_standalone_model:
      self._run_returnn_standalone_net_dict()
      self._run_returnn_standalone_python()

  @property
  def returnn_net_dict(self) -> Dict[str, Dict[str, Any]]:
    assert self._returnn_net_dict, "Call run() first."
    return self._returnn_net_dict

  def _make_tf_feed_dict(self, input: Data):
    assert input.batch_ndim == len(self._inputs_np.shape)
    assert all(input.batch_shape[i] in {None, self._inputs_np.shape[i]} for i in range(input.batch_ndim))
    n_batch = 1 if input.batch_dim_axis is None else self._inputs_np.shape[input.batch_dim_axis]
    d = {input.placeholder: self._inputs_np}
    if self._returnn_input_seq_lens is not None:
      assert tuple(self._returnn_input_seq_lens.keys()) == tuple(input.size_placeholder.keys())
      for i, size in input.size_placeholder.items():
        d[size] = self._returnn_input_seq_lens[i]
    else:
      for i, size in input.size_placeholder.items():
        d[size] = [self._inputs_np.shape[input.get_batch_axis(i)]] * n_batch  # not so relevant
    return d

  def _run_reference(self):
    """
    The reference, using the original import.

    This is only needed if you want to have this as the reference output,
    for further verification checks.
    Otherwise it can be skipped.
    We can use the outputs from the traced Torch variant
    (:func:`_run_traced_orig_torch`) instead.
    """
    print(">>> Running with standard reference imports...")
    torch.manual_seed(42)
    numpy.random.seed(42)
    with torch.no_grad():
      out_ref = self._model_func(None, torch.from_numpy(self._inputs_np.copy()))
      assert isinstance(out_ref, torch.Tensor)
      out_ref_np = out_ref.cpu().numpy()
    self._out_ref_np = out_ref_np
    print()

  def _run_traced_orig_torch(self):
    """
    Now with wrapped import. That will also use the original PyTorch code, but wrapped with our custom logic.
    This should not change anything, and still would use the PyTorch logic,
    except that the wrapped classes can collect additional information.
    However, we still will check that we got the same output,
    just to check that there is no subtle bug due to the wrapping logic.
    """
    print(">>> Running with wrapped imports, wrapping original PyTorch...")
    torch.manual_seed(42)
    numpy.random.seed(42)
    with torch.no_grad():
      with Naming.make_instance(
            wrap_to_returnn_enabled=False,
            keep_orig_module_io_tensors=self.verify_individual_model_io) as naming:
        wrapped_torch = wrapped_import_torch_traced("torch")
        out_wrapped = self._model_func(wrapped_import_torch_traced, wrapped_torch.from_numpy(self._inputs_np.copy()))
        assert isinstance(out_wrapped, WrappedTorchTensor)
        out_wrapped_np = out_wrapped.cpu().numpy()
        print(">>>> Module naming hierarchy:")
        naming.root_namespace.dump()
        print(">>>> Root module calls:")
        pprint(dict(naming.get_root_module_calls()))
        torch_mods_with_params = naming.get_modules_with_params_by_abs_name()
        print(">>>> Modules with params:")
        pprint(dict(torch_mods_with_params))
        self._torch_namespace = naming
    if self._out_ref_np is not None:
      assert self._out_ref_np.shape == out_wrapped_np.shape
      numpy.testing.assert_allclose(self._out_ref_np, out_wrapped_np)
      print(">>>> Looks good!")
    else:
      self._out_ref_np = out_wrapped_np  # just use that as further reference
    print()

  def _run_torch_returnn_drop_in(self):
    print(">>> Running with wrapped Torch import, wrapping replacement for PyTorch...")
    torch.manual_seed(42)
    numpy.random.seed(42)
    with make_scope() as session:
      with Naming.make_instance(
            wrap_to_returnn_enabled=True,
            returnn_train_flag=self.train,
            keep_orig_module_io_tensors=True,  # it's only symbolic anyway in TF
            import_params_from_torch_namespace=self._torch_namespace) as naming:
        assert isinstance(naming, Naming)
        in_returnn = torch_returnn.from_numpy(self._inputs_np.copy())
        assert isinstance(in_returnn, torch_returnn.Tensor)
        x = naming.register_input(in_returnn, Data("data", **self._returnn_in_data_dict))
        print("RETURNN input:", x)
        out_returnn = self._model_func(wrapped_import_torch_returnn, in_returnn)
        assert isinstance(out_returnn, torch_returnn.Tensor)
        out_returnn_ = naming.register_output(out_returnn)
        y, returnn_axis_from_torch_axis = out_returnn_.returnn_data, out_returnn_.returnn_axis_from_torch_axis
        assert isinstance(y, Data)
        print("RETURNN output:", y, "axis map RETURNN<-Torch", returnn_axis_from_torch_axis)
        print(">>>> Module naming hierarchy:")
        naming.root_namespace.dump()
        print(">>>> RETURNN net dict:")
        self._returnn_net_dict = naming.root_namespace.dump_as_returnn_net_dict()
        pprint(self._returnn_net_dict)
        print(">>>> Root module calls:")
        pprint(dict(naming.get_root_module_calls()))
        torch_mods_with_params = naming.get_modules_with_params_by_abs_name()
        print(">>>> Modules with params:")
        pprint(dict(torch_mods_with_params))

      feed_dict = self._make_tf_feed_dict(x)
      y_, y_size = session.run((y.placeholder, y.size_placeholder.as_dict()), feed_dict=feed_dict)
      assert isinstance(y_, numpy.ndarray)
      self._out_returnn_np = y_
      print("Output shape:", y_.shape)
      print("Output seq lens:", y_size)
      y_torch = y_.transpose(*[returnn_axis_from_torch_axis[i] for i in range(y_.ndim)])
      print("Output shape (converted to Torch):", y_torch.shape)
      if self._out_ref_np is not None:
        if self._returnn_input_seq_lens is not None and y_size:
          # ignore padded elements in comparison
          assert tuple(y_size.keys()) == tuple(self._returnn_input_seq_lens.keys())
          assert self._out_ref_np.shape == y_torch.shape, "Mismatching shapes"
          for ax in y.get_spatial_axes():
            assert y_size[0].tolist() == self._returnn_input_seq_lens[0], (
              "Mismatch of sequence lens, not yet implemented")
            torch_axis_from_returnn_axis = {j: i for i, j in returnn_axis_from_torch_axis.items()}
            batch_dim_axis = torch_axis_from_returnn_axis[y.batch_dim_axis]
            spatial_axis = torch_axis_from_returnn_axis[y.get_batch_axis(ax)]
            for seq_idx, seq_len in enumerate(y_size[ax]):
              slc = [slice(None)] * len(y_torch.shape)
              slc[batch_dim_axis] = seq_idx
              slc[spatial_axis] = slice(seq_len, y_torch.shape[spatial_axis])
              self._out_ref_np[tuple(slc)] = numpy.nan
        m = ~numpy.isnan(self._out_ref_np)
        numpy.testing.assert_allclose(self._out_ref_np[m], y_torch[m], **naming.validate_allclose_kwargs)
        print(">>>> Looks good!")

      if self.export_tf_checkpoint_save_path or self.verify_returnn_standalone_model:
        returnn_net = naming.root_namespace.returnn_ctx.network
        returnn_net.print_network_info(name="RETURNN network")
        if self.export_tf_checkpoint_save_path:
          self._tf_checkpoint_save_path = self.export_tf_checkpoint_save_path
        else:
          tmp_dir = tempfile.mkdtemp("tmp-returnn-tf-checkpoint")
          self._tf_checkpoint_save_path = tmp_dir + "/model"
        print(f"Saving TF checkpoint to {self._tf_checkpoint_save_path!r}...")
        returnn_net.global_train_step.load(0, session=session)
        returnn_net.save_params_to_file(filename=self._tf_checkpoint_save_path, session=session)
        print()

  def _run_returnn_standalone_net_dict(self):
    print(">>> Constructing RETURNN model, load TF checkpoint, run...")
    with make_scope() as session:
      from returnn.config import Config
      from returnn.tf.network import TFNetwork
      config = Config({
        "extern_data": {"data": self._returnn_in_data_dict},
        "debug_print_layer_output_template": True,
      })
      network = TFNetwork(config=config, name="root", train_flag=self.train)
      network.construct_from_dict(self._returnn_net_dict)
      network.load_params_from_file(filename=self._tf_checkpoint_save_path, session=session)

      x = network.extern_data.get_default_input_data()
      y = network.get_default_output_layer().output
      feed_dict = self._make_tf_feed_dict(x)
      y_, y_size = session.run((y.placeholder, y.size_placeholder.as_dict()), feed_dict=feed_dict)
      assert isinstance(y_, numpy.ndarray)
      print("Output shape:", y_.shape)
      numpy.testing.assert_allclose(self._out_returnn_np, y_)
      print(">>>> Looks good!")
      print()

  def _run_returnn_standalone_python(self):
    print(">>> Constructing RETURNN model via Python code, load TF checkpoint, run...")
    with make_scope() as session:
      with Naming.make_instance() as naming:  # we expect this to work with the default settings
        model_func = self._model_func

        # Wrap the model_func in a module.
        # We assume this would be flattened away in the namespace.
        # All named modules should thus have the same names.
        class DummyModule(torch_returnn.nn.Module):
          returnn_dummy_input_shape = self._returnn_dummy_input_shape

          def get_returnn_name(self) -> str:
            return ""  # also avoid that this name becomes a prefix anywhere

          def forward(self, *inputs):
            return model_func(wrapped_import_torch_returnn, *inputs)

        dummy_mod = DummyModule()
        net_dict = dummy_mod.as_returnn_net_dict(self._returnn_in_data_dict)

      from returnn.config import Config
      from returnn.tf.network import TFNetwork
      config = Config({
        "extern_data": {"data": self._returnn_in_data_dict},
        "debug_print_layer_output_template": True,
      })
      network = TFNetwork(config=config, name="root", train_flag=self.train)
      network.construct_from_dict(net_dict)
      network.load_params_from_file(filename=self._tf_checkpoint_save_path, session=session)

      x = network.extern_data.get_default_input_data()
      y = network.get_default_output_layer().output
      feed_dict = self._make_tf_feed_dict(x)
      y_, y_size = session.run((y.placeholder, y.size_placeholder.as_dict()), feed_dict=feed_dict)
      assert isinstance(y_, numpy.ndarray)
      print("Output shape:", y_.shape)
      numpy.testing.assert_allclose(self._out_returnn_np, y_)
      print(">>>> Looks good!")
      print()


def verify_torch_and_convert_to_returnn(
      model_func: ModelFuncType, *,
      inputs: numpy.ndarray, **kwargs) -> Converter:
  """
  :param model_func:
    Gets an argument wrapped_import(str) -> module, or None. If None, should import as is.
    It also gets the inputs, converted to the right PyTorch `Tensor` object (either original or wrapped).
  :param inputs:

  model_func will get called multiple times, with different wrapped_import functions.
  wrapped_import would import some user model code.
  wrapped_import expects that the user model code is still unmodified,
  using the original `import torch` statements.

  See the `readme <..>`_ for further details.
  """
  converter = Converter(model_func=model_func, inputs=inputs, **kwargs)
  converter.run()
  return converter




"""
This is used for parameter initialization.

We currently mostly ignore this.
In principle, we could keep track of this,
and then set corresponding parameter init options on layers in RETURNN,
such as ``forward_weights_init``, etc.
However, even if we do, we cannot really verify whether it is correct.
"""

from ..tensor import Tensor


def _calculate_fan_in_and_fan_out(tensor: Tensor):
  dimensions = tensor.dim()
  if dimensions < 2:
    raise ValueError("Fan in and fan out can not be computed for tensor with fewer than 2 dimensions")

  num_input_fmaps = tensor.size(1)
  num_output_fmaps = tensor.size(0)
  receptive_field_size = 1
  if tensor.dim() > 2:
    for s in tensor.shape[2:]:
      receptive_field_size *= s
  fan_in = num_input_fmaps * receptive_field_size
  fan_out = num_output_fmaps * receptive_field_size

  return fan_in, fan_out


def zeros_(tensor: Tensor):
  tensor.zero_()


def ones_(tensor: Tensor):
  tensor.fill_(1)


def constant_(tensor: Tensor, value=0):
  tensor.fill_(value)


def uniform_(tensor, a=0., b=1.):
  pass


def normal_(tensor, mean=0., std=1.):
  pass


def xavier_uniform_(tensor: Tensor, gain=1.):
  pass


def xavier_normal_(tensor: Tensor, gain=1.):
  pass


def kaiming_uniform_(tensor: Tensor, a=0, mode='fan_in', nonlinearity='leaky_relu'):
  pass


def kaiming_normal_(tensor: Tensor, a=0, mode='fan_in', nonlinearity='leaky_relu'):
  pass




from __future__ import annotations
import tensorflow as tf
import math
from typing import Optional, Dict, List, Tuple, Any
from returnn.tf.layers.basic import LayerBase, ConvLayer
from .module import Module
from .utils import _single, _pair, _triple, _reverse_repeat_tuple, _ntuple
from ..common_types import _scalar_or_tuple_any_t, _size_1_t, _size_2_t, _size_3_t
from ...tensor import Tensor
from ..parameter import Parameter
from .. import init


class _ConvNd(Module):
  nd: Optional[int] = None  # defined by subclass
  transposed: bool = False

  def __init__(self,
               in_channels: int,
               out_channels: int,
               kernel_size: _scalar_or_tuple_any_t,
               stride: _scalar_or_tuple_any_t = 1,
               padding: _scalar_or_tuple_any_t = 0,
               dilation: _scalar_or_tuple_any_t = 1,
               transposed: Optional[bool] = None,
               output_padding: _scalar_or_tuple_any_t = 0,
               groups: int = 1,
               bias: bool = True,
               padding_mode: str = "zeros") -> None:
    super(_ConvNd, self).__init__()
    if in_channels % groups != 0:
      raise ValueError('in_channels must be divisible by groups')
    if out_channels % groups != 0:
      raise ValueError('out_channels must be divisible by groups')
    valid_padding_modes = {'zeros', 'reflect', 'replicate', 'circular'}
    if padding_mode not in valid_padding_modes:
      raise ValueError("padding_mode must be one of {}, but got padding_mode='{}'".format(
        valid_padding_modes, padding_mode))
    self.in_channels = in_channels
    self.out_channels = out_channels
    self.kernel_size = _ntuple(self.nd)(kernel_size)
    self.stride = _ntuple(self.nd)(stride)
    self.padding = _ntuple(self.nd)(padding)
    self.dilation = _ntuple(self.nd)(dilation)
    if transposed is not None:
      self.transposed = transposed
    self.output_padding = _ntuple(self.nd)(output_padding)
    self.groups = groups
    self.padding_mode = padding_mode
    # `_reversed_padding_repeated_twice` is the padding to be passed to
    # `F.pad` if needed (e.g., for non-zero padding types that are
    # implemented as two ops: padding + conv). `F.pad` accepts paddings in
    # reverse order than the dimension.
    self._reversed_padding_repeated_twice = _reverse_repeat_tuple(self.padding, 2)
    if self.transposed:
      self.weight = Parameter(Tensor(
        in_channels, out_channels // groups, *self.kernel_size))
    else:
      self.weight = Parameter(Tensor(
        out_channels, in_channels // groups, *self.kernel_size))
    if bias:
      self.bias = Parameter(Tensor(out_channels))
    else:
      self.register_parameter('bias', None)
    self.reset_parameters()

  def reset_parameters(self) -> None:
    init.kaiming_uniform_(self.weight, a=math.sqrt(5))
    if self.bias is not None:
      fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
      bound = 1 / math.sqrt(fan_in)
      init.uniform_(self.bias, -bound, bound)

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    assert len(input.shape) == 2 + self.nd
    self._assert_spatial_axes_in_order(input)  # not implemented otherwise
    assert all(p == 0 for p in self.padding)  # not implemented otherwise
    assert all(p == 0 for p in self.output_padding)  # not implemented otherwise
    assert self.padding_mode == "zeros"  # not implemented otherwise
    d = {
      "class": "conv", "from": self._get_input_layer_name(input),
      "activation": None,
      "with_bias": self.bias is not None,
      "n_out": self.out_channels,
      "filter_size": self.kernel_size,
      "padding": "valid"}
    if any(s != 1 for s in self.stride):
      d["strides"] = self.stride
    if any(d != 1 for d in self.dilation):
      d["dilation_rate"] = self.dilation
    if self.groups != 1:
      d["groups"] = self.groups
    return d

  def check_returnn_layer(self, layer: ConvLayer):
    assert layer.input_data.dim == self.in_channels

  def import_params_torch_to_returnn(self, *, layer: LayerBase, torch_module: _ConvNd):
    session = tf.compat.v1.get_default_session()
    if self.transposed:
      # E.g. convert 11,13,10 -> 10,1,13,11.
      values = torch_module.weight.detach().numpy()
      assert len(self.kernel_size) in {1, 2}
      axes = list(range(values.ndim))
      values = values.transpose(*(axes[2:] + [axes[1], axes[0]]))
      if len(self.kernel_size) == 1:
        values = values[:, None]
      layer.params["W_native_transposed_conv"].load(values, session=session)
    else:
      # E.g. 384,80,7 -> 7,80,384
      values = torch_module.weight.detach().numpy()
      axes = list(range(values.ndim))
      values = values.transpose(*(axes[2:] + [axes[1], axes[0]]))
      layer.params["W"].load(values, session=session)
    if self.bias is not None:
      layer.params["bias"].load(torch_module.bias.detach().numpy(), session=session)

  def _get_output_shape_from_returnn(self, inputs_flat: List[Tensor], layer: LayerBase
                                     ) -> Tuple[Tuple[int, ...], Dict[int, int]]:
    """
    The basic returnn_axis_from_torch_axis should be correct, however, if the size of a dynamic axis changes (e.g. due
    to strides and/or padding), this is not covered in the base method and we fix it here.
    """
    torch_shape, returnn_axis_from_torch_axis = super(_ConvNd, self)._get_output_shape_from_returnn(
      inputs_flat=inputs_flat, layer=layer)
    assert len(inputs_flat) == 1
    torch_shape = list(inputs_flat[0].shape)
    torch_shape[1] = self.out_channels
    for idx in range(self.nd):
      torch_ax = idx + 2
      torch_shape[torch_ax] = (torch_shape[torch_ax] + 2 * self.padding[idx] - self.dilation[idx] * (
        self.kernel_size[idx] - 1) - 1) // self.stride[idx] + 1
    return tuple(torch_shape), returnn_axis_from_torch_axis


class Conv1d(_ConvNd):
  nd = 1

  def __init__(
      self,
      in_channels: int,
      out_channels: int,
      kernel_size: _size_1_t,
      stride: _size_1_t = 1,
      padding: _size_1_t = 0,
      dilation: _size_1_t = 1,
      groups: int = 1,
      bias: bool = True,
      padding_mode: str = 'zeros'
  ):
    super(Conv1d, self).__init__(
      in_channels, out_channels, kernel_size, stride,
      padding=padding, dilation=dilation,
      groups=groups, bias=bias, padding_mode=padding_mode)


class Conv2d(_ConvNd):
  nd = 2

  def __init__(
      self,
      in_channels: int,
      out_channels: int,
      kernel_size: _size_2_t,
      stride: _size_2_t = 1,
      padding: _size_2_t = 0,
      dilation: _size_2_t = 1,
      groups: int = 1,
      bias: bool = True,
      padding_mode: str = 'zeros'
  ):
    super(Conv2d, self).__init__(
      in_channels, out_channels, kernel_size, stride,
      padding=padding, dilation=dilation,
      groups=groups, bias=bias, padding_mode=padding_mode)


class _ConvTransposeNd(_ConvNd):
  def __init__(self, in_channels, out_channels, kernel_size, stride,
               padding, dilation, transposed, output_padding,
               groups, bias, padding_mode):
    if padding_mode != 'zeros':
      raise ValueError('Only "zeros" padding mode is supported for {}'.format(self.__class__.__name__))

    super(_ConvTransposeNd, self).__init__(
      in_channels, out_channels, kernel_size, stride,
      padding, dilation, transposed, output_padding,
      groups, bias, padding_mode)

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    assert len(input.shape) == 2 + self.nd
    self._assert_spatial_axes_in_order(input)  # not implemented otherwise
    assert self.groups == 1  # not implemented otherwise
    assert self.padding_mode == "zeros"  # not implemented otherwise
    assert all(d == 1 for d in self.dilation)
    d = {
      "class": "transposed_conv", "from": self._get_input_layer_name(input),
      "activation": None,
      "with_bias": self.bias is not None,
      "n_out": self.out_channels,
      "filter_size": self.kernel_size,
      "strides": self.stride,
      "padding": "valid",
      "output_padding": self.output_padding}
    if self.padding:
      d["remove_padding"] = self.padding
    return d

  def _get_output_shape_from_returnn(self, inputs_flat: List[Tensor], layer: LayerBase
                                     ) -> Tuple[Tuple[int, ...], Dict[int, int]]:
    """
    The basic returnn_axis_from_torch_axis should be correct, however, if the size of a dynamic axis changes (e.g. due
    to strides and/or padding), this is not covered in the base method and we fix it here.
    """
    torch_shape, returnn_axis_from_torch_axis = super(_ConvNd, self)._get_output_shape_from_returnn(
      inputs_flat=inputs_flat, layer=layer)
    assert len(inputs_flat) == 1
    torch_shape = list(inputs_flat[0].shape)
    torch_shape[1] = self.out_channels
    for idx in range(self.nd):
      torch_ax = idx + 2
      torch_shape[torch_ax] = (torch_shape[torch_ax] - 1) * self.stride[idx] - 2 * self.padding[idx] + (
        self.dilation[idx] * (self.kernel_size[idx] - 1) + self.output_padding[idx] + 1)
    return tuple(torch_shape), returnn_axis_from_torch_axis


class ConvTranspose1d(_ConvTransposeNd):
  nd = 1

  def __init__(
      self,
      in_channels: int,
      out_channels: int,
      kernel_size: _size_1_t,
      stride: _size_1_t = 1,
      padding: _size_1_t = 0,
      output_padding: _size_1_t = 0,
      groups: int = 1,
      bias: bool = True,
      dilation: _size_1_t = 1,
      padding_mode: str = 'zeros'
  ):
    super(ConvTranspose1d, self).__init__(
      in_channels, out_channels, kernel_size, stride, padding, dilation,
      True, output_padding, groups, bias, padding_mode)


class ConvTranspose2d(_ConvTransposeNd):
  nd = 2

  def __init__(
      self,
      in_channels: int,
      out_channels: int,
      kernel_size: _size_2_t,
      stride: _size_2_t = 1,
      padding: _size_2_t = 0,
      output_padding: _size_2_t = 0,
      groups: int = 1,
      bias: bool = True,
      dilation: _size_2_t = 1,
      padding_mode: str = 'zeros'
  ):
    super(ConvTranspose2d, self).__init__(
      in_channels, out_channels, kernel_size, stride, padding, dilation,
      True, output_padding, groups, bias, padding_mode)


class _FunctionalConvNd(Module):
  is_original_torch_module = False
  nd: Optional[int] = None
  transposed: bool = False

  def __init__(self,
               stride: _scalar_or_tuple_any_t = 1,
               padding: _scalar_or_tuple_any_t = 0,
               dilation: _scalar_or_tuple_any_t = 1,
               output_padding: _scalar_or_tuple_any_t = 0,
               groups: int = 1,
               padding_mode: str = "zeros") -> None:
    super(_FunctionalConvNd, self).__init__()
    valid_padding_modes = {'zeros', 'reflect', 'replicate', 'circular'}
    if padding_mode not in valid_padding_modes:
      raise ValueError("padding_mode must be one of {}, but got padding_mode='{}'".format(
        valid_padding_modes, padding_mode))
    self.stride = _ntuple(self.nd)(stride)
    self.padding = _ntuple(self.nd)(padding)
    self.dilation = _ntuple(self.nd)(dilation)
    self.output_padding = _ntuple(self.nd)(output_padding)
    self.groups = groups
    self.padding_mode = padding_mode
    # `_reversed_padding_repeated_twice` is the padding to be passed to
    # `F.pad` if needed (e.g., for non-zero padding types that are
    # implemented as two ops: padding + conv). `F.pad` accepts paddings in
    # reverse order than the dimension.
    self._reversed_padding_repeated_twice = _reverse_repeat_tuple(self.padding, 2)

  def create_returnn_layer_dict(self, input: Tensor, weight: Tensor, bias: Optional[Tensor] = None) -> Dict[str, Any]:
    assert len(input.shape) == 2 + self.nd
    self._assert_spatial_axes_in_order(input)  # not implemented otherwise
    assert len(weight.shape) == 2 + self.nd
    kernel_size = weight.shape[2:]
    out_channels, in_channels = weight.shape[:2]
    in_channels *= self.groups
    if self.transposed:
      out_channels, in_channels = in_channels, out_channels
    returnn_weight_axes = [self._get_input_axis_to_returnn(weight, i) for i in range(len(weight.shape))]
    returnn_weight_axes_ = returnn_weight_axes[2:] + [returnn_weight_axes[1], returnn_weight_axes[0]]
    returnn_weight_transpose_perm = {i: j for (i, j) in zip(returnn_weight_axes, returnn_weight_axes_)}
    if bias is not None:
      assert bias.shape == (out_channels,)
    assert self.groups == 1  # not implemented otherwise
    assert self.padding_mode == "zeros"  # not implemented otherwise
    if self.transposed:  # transposed conv
      assert all(d == 1 for d in self.dilation)
      return {
        "class": "transposed_conv", "from": self._get_input_layer_name(input),
        "n_out": out_channels,
        "activation": None,
        "with_bias": bias is not None,
        "bias": self._get_input_layer_name(bias) if bias is not None else None,
        "filter_size": kernel_size,
        "filter": self._get_input_layer_name(weight),
        "filter_perm": returnn_weight_transpose_perm,
        "padding": "valid",
        "output_padding": self.output_padding,
        "remove_padding": self.padding,
        "strides": self.stride}
    else:
      assert all(p == 0 for p in self.padding)  # not implemented otherwise
      assert all(p == 0 for p in self.output_padding)  # not implemented otherwise
      return {
        "class": "conv", "from": self._get_input_layer_name(input),
        "n_out": out_channels,
        "activation": None,
        "with_bias": bias is not None,
        "bias": self._get_input_layer_name(bias) if bias is not None else None,
        "filter_size": kernel_size,
        "filter": self._get_input_layer_name(weight),
        "filter_perm": returnn_weight_transpose_perm,
        "padding": "valid",
        "strides": self.stride,
        "dilation_rate": self.dilation}


class FunctionalConv1d(_FunctionalConvNd):
  nd = 1


class FunctionalConv2d(_FunctionalConvNd):
  nd = 2


class FunctionalConvTransposed1d(_FunctionalConvNd):
  nd = 1
  transposed = True


__all__ = [
  key for (key, value) in sorted(globals().items())
  if not key.startswith("_")
  and getattr(value, "__module__", "") == __name__]



###############################################################################
# Language Modeling on Wikitext-2
#
# This file generates new sentences sampled from the language model
#
###############################################################################

import argparse

import torch

import data

parser = argparse.ArgumentParser(description='PyTorch Wikitext-2 Language Model')

# Model parameters.
parser.add_argument('--data', type=str, default='./data/wikitext-2',
                    help='location of the data corpus')
parser.add_argument('--checkpoint', type=str, default='./model.pt',
                    help='model checkpoint to use')
parser.add_argument('--outf', type=str, default='generated.txt',
                    help='output file for generated text')
parser.add_argument('--words', type=int, default='1000',
                    help='number of words to generate')
parser.add_argument('--seed', type=int, default=1111,
                    help='random seed')
parser.add_argument('--cuda', action='store_true',
                    help='use CUDA')
parser.add_argument('--temperature', type=float, default=1.0,
                    help='temperature - higher will increase diversity')
parser.add_argument('--log-interval', type=int, default=100,
                    help='reporting interval')
args = parser.parse_args()

# Set the random seed manually for reproducibility.
torch.manual_seed(args.seed)
if torch.cuda.is_available():
    if not args.cuda:
        print("WARNING: You have a CUDA device, so you should probably run with --cuda")

device = torch.device("cuda" if args.cuda else "cpu")

if args.temperature < 1e-3:
    parser.error("--temperature has to be greater or equal 1e-3")

with open(args.checkpoint, 'rb') as f:
    model = torch.load(f).to(device)
model.eval()

corpus = data.Corpus(args.data)
ntokens = len(corpus.dictionary)

is_transformer_model = hasattr(model, 'model_type') and model.model_type == 'Transformer'
if not is_transformer_model:
    hidden = model.init_hidden(1)
input = torch.randint(ntokens, (1, 1), dtype=torch.long).to(device)

with open(args.outf, 'w') as outf:
    with torch.no_grad():  # no tracking history
        for i in range(args.words):
            if is_transformer_model:
                output = model(input, False)
                word_weights = output[-1].squeeze().div(args.temperature).exp().cpu()
                word_idx = torch.multinomial(word_weights, 1)[0]
                word_tensor = torch.Tensor([[word_idx]]).long().to(device)
                input = torch.cat([input, word_tensor], 0)
            else:
                output, hidden = model(input, hidden)
                word_weights = output.squeeze().div(args.temperature).exp().cpu()
                word_idx = torch.multinomial(word_weights, 1)[0]
                input.fill_(word_idx)

            word = corpus.dictionary.idx2word[word_idx]

            outf.write(word + ('\n' if i % 20 == 19 else ' '))

            if i % args.log_interval == 0:
                print('| Generated {}/{} words'.format(i, args.words))




from __future__ import annotations
import math
import numbers
from typing import Dict, Any, Union, List
import tensorflow as tf
from returnn.tf.layers.basic import LayerNormLayer, NormLayer
from .module import Module
from ..parameter import Parameter
from ...tensor import Tensor, Size
from .. import init


_shape_t = Union[int, List[int], Size]


class _Norm(Module):
  def __init__(self, axes: str, normalized_shape: _shape_t = None, eps: float = 1e-5, affine: bool = True) -> None:
    super(_Norm, self).__init__()
    self.axes = axes
    self.eps = eps
    self.affine = affine
    if self.affine:
      self.weight = Parameter(Tensor(normalized_shape))
      self.bias = Parameter(Tensor(normalized_shape))
    else:
      self.register_parameter('weight', None)
      self.register_parameter('bias', None)
    self.reset_parameters()

  def reset_parameters(self) -> None:
    if self.affine:
      init.ones_(self.weight)
      init.zeros_(self.bias)

  def import_params_torch_to_returnn(self, *, layer: NormLayer, torch_module: GroupNorm):
    assert isinstance(layer, NormLayer)
    if self.affine:
      session = tf.compat.v1.get_default_session()
      layer.params["scale"].load(torch_module.weight.detach().numpy(), session=session)
      layer.params["bias"].load(torch_module.bias.detach().numpy(), session=session)

  def create_returnn_layer_dict(self, input: Tensor):
    return {
      "class": "norm", "axes": self.axes, "epsilon": self.eps, "scale": self.affine, "bias": self.affine,
      "from": self._get_input_layer_name(input)}


class LayerNorm(_Norm):
  def __init__(self, normalized_shape: _shape_t, eps: float = 1e-5, elementwise_affine: bool = True) -> None:
    if isinstance(normalized_shape, numbers.Integral):
      normalized_shape = (normalized_shape,)
    self.normalized_shape = tuple(normalized_shape)
    self.elementwise_affine = elementwise_affine
    super(LayerNorm, self).__init__(axes="F", normalized_shape=normalized_shape, eps=eps, affine=elementwise_affine)


class GroupNorm(_Norm):
  def __init__(self, num_groups: int, num_channels: int, eps: float = 1e-5, affine: bool = True) -> None:
    assert num_groups == 1 or num_groups == num_channels, "Not implemented otherwise"
    self.num_groups = num_groups
    self.num_channels = num_channels

    axes = None
    if num_groups == 1:
      axes = "TF"
    elif num_groups == num_channels:
      axes = "T"
    super(GroupNorm, self).__init__(axes=axes, normalized_shape=num_channels, eps=eps, affine=affine)


__all__ = [
  key for (key, value) in sorted(globals().items())
  if not key.startswith("_")
  and getattr(value, "__module__", "") == __name__]





def load(*args, **kwargs):
  # This is just unpickling. It's okay if we get some torch.Tensor or torch.nn.Parameter in the dict.
  # Our code in Module.load_state_dict takes care of that.
  import torch
  return torch.load(*args, **kwargs)


def default_restore_location():
  """Dummy function to avoid import errors"""
  pass




import functools
import inspect
from typing import TypeVar, Callable, Any, cast


FuncType = Callable[..., Any]
F = TypeVar('F', bound=FuncType)


class _DecoratorContextManager:
  """Allow a context manager to be used as a decorator"""

  def __call__(self, func: F) -> F:
    if inspect.isgeneratorfunction(func):
      return self._wrap_generator(func)

    @functools.wraps(func)
    def decorate_context(*args, **kwargs):
      with self.__class__():
        return func(*args, **kwargs)

    return cast(F, decorate_context)

  def _wrap_generator(self, func):
    """Wrap each generator invocation with the context manager"""

    @functools.wraps(func)
    def generator_context(*args, **kwargs):
      gen = func(*args, **kwargs)
      while True:
        try:
          with self.__class__():
            x = next(gen)
          yield x
        except StopIteration:
          break

    return generator_context

  def __enter__(self) -> None:
    raise NotImplementedError

  def __exit__(self, exc_type, exc_value, traceback):
    raise NotImplementedError


class no_grad(_DecoratorContextManager):
  def __enter__(self):
    pass

  def __exit__(self, exc_type, exc_value, traceback):
    pass




from typing import List
from .module import Module
from ...tensor import Tensor


class Norm(Module):
  is_original_torch_module = False

  def __init__(self, *, axes: List[int], p: float = 2, keepdims: bool = False):
    super(Norm, self).__init__()
    self.axes = axes
    self.p = p
    self.keepdims = keepdims

  def create_returnn_layer_dict(self, input: Tensor):
    return {
      "class": "math_norm", "p": self.p, "keep_dims": self.keepdims,
      "axes": [self._get_input_axis_to_returnn(input, axis) for axis in self.axes],
      "from": self._get_input_layer_name(input)}


__all__ = [
  key for (key, value) in sorted(globals().items())
  if not key.startswith("_")
  and getattr(value, "__module__", "") == __name__]




import sys
from typing import Any
import importlib
from ... import log
from ... import __package__ as _base_package
from ..context import WrapCtx


class WrappedObject:
  def __init__(self, orig_obj: Any, name: str, ctx: WrapCtx):
    self._wrapped__name = name
    self._wrapped__orig_obj = orig_obj
    self._wrapped__ctx = ctx

  def __repr__(self):
    if self._wrapped__orig_obj is self:
      return super(WrappedObject, self).__repr__()
    return "<WrappedObject %r type %s>" % (self._wrapped__name, type(self._wrapped__orig_obj))

  def __getattr__(self, item):
    if item == "_wrapped__orig_obj":
      # Special case. torch.Tensor functions would create new copies,
      # which would use our custom class, but then this is not initialized.
      return self
    if item == "_wrapped__name":
      return "<unknown>"
    if item == "_wrapped__ctx":
      return None
    if self._wrapped__orig_obj is self:  # special case
      raise AttributeError("No attrib %r" % item)
    from ..wrap import wrap
    res_ = getattr(self._wrapped__orig_obj, item)
    try:
      res = wrap(res_, name="%s.%s" % (self._wrapped__name, item), ctx=self._wrapped__ctx)
      if res is not res_:
        # Speedup, and avoid recreations.
        # Note that this can potentially be dangerous and lead to inconsistencies,
        # esp if the underlying object changes the value,
        # then this would not reflect that.
        # TODO maybe remove this?
        object.__setattr__(self, item, res)
      else:
        if log.Verbosity >= 10:
          log.unique_print("*** not wrapped: '%s.%s', type %r" % (self._wrapped__name, item, type(res)))
      if log.Verbosity >= 6:
        postfix = ""
        if type(res).__name__.startswith("Wrapped") or getattr(res, "__module__", "").startswith(_base_package):
          postfix = " -> %r" % res
        log.unique_print("*** indirect getattr '%s.%s'%s" % (self._wrapped__name, item, postfix))
    except AttributeError as exc:  # no exception expected. esp, we should **NOT** forward AttributeError
      raise RuntimeError(exc) from exc
    return res

  # setattr on this object directly, not on the wrapped one
  # Note again that this can potentially lead to inconsistencies. See note above.
  __setattr__ = object.__setattr__

  def __delattr__(self, item):
    if hasattr(self._wrapped__orig_obj, item):
      delattr(self._wrapped__orig_obj, item)
    if item in self.__dict__:
      del self.__dict__[item]

  def __dir__(self):
    if self._wrapped__orig_obj is self:
      return []
    return dir(self._wrapped__orig_obj)

  def __bool__(self):
    if self._wrapped__orig_obj is self:
      # noinspection PyUnresolvedReferences
      return super(WrappedObject, self).__bool__(self)
    return bool(self._wrapped__orig_obj)


def make_wrapped_object(obj, *, name: str, ctx: WrapCtx):
  # First check if we should also wrap the type.
  if type(obj) in ctx.explicit_wrapped_types:
    wrap_type = ctx.explicit_wrapped_types[type(obj)]
    return wrap_type.wrap(obj, name=name, ctx=ctx)

  if ctx.should_wrap_mod(getattr(type(obj), "__module__", "")) and type(obj) not in ctx.keep_as_is_types:
    from .class_ import WrappedClassBase
    assert getattr(sys.modules[obj.__class__.__module__], obj.__class__.__qualname__) is obj.__class__
    wrapped_mod = importlib.import_module(ctx.wrapped_mod_prefix + obj.__class__.__module__)
    cls = getattr(wrapped_mod, obj.__class__.__qualname__)
    assert issubclass(cls, WrappedClassBase)
    obj = cls(orig_obj=obj, name=name, ctx=ctx)
    return obj

  # Fallback, or standard case.
  return WrappedObject(obj, name=name, ctx=ctx)




from typing import Dict, Any
from .module import Module
from ...tensor import Tensor


class _DropoutNd(Module):
  p: float
  inplace: bool

  def __init__(self, p: float = 0.5, inplace: bool = False) -> None:
    super(_DropoutNd, self).__init__()
    if p < 0 or p > 1:
      raise ValueError("dropout probability has to be between 0 and 1, "
                       "but got {}".format(p))
    self.p = p
    self.inplace = inplace  # just ignore...


class Dropout(_DropoutNd):
  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    return {"class": "dropout", "dropout": self.p, "from": self._get_input_layer_name(input)}


__all__ = [
  key for (key, value) in sorted(globals().items())
  if not key.startswith("_")
  and getattr(value, "__module__", "") == __name__]




from __future__ import annotations
import numpy
import tensorflow as tf
from returnn.tf.layers.basic import LayerBase, VariableLayer
from typing import Tuple, List, Dict, Optional
from ...tensor import Tensor
from ..parameter import Parameter
from .module import Module
from ....naming import Naming


class Variable(Module):
  is_original_torch_module = False

  def __init__(self, param: Parameter, parent_mod: Optional[Tuple[Module, str]] = None):
    super(Variable, self).__init__()
    assert isinstance(param, Parameter)
    self.param = param
    self.parent_mod = parent_mod

  def create_returnn_layer_dict(self, *inputs):  # ignore inputs
    return {"class": "variable", "add_batch_axis": False, "shape": self.param.shape}

  def make_output_tensor_from_returnn(self, inputs_flat: List[Tensor], layer: LayerBase) -> Tensor:
    naming = Naming.get_instance()
    values = None
    if naming.import_params_from_torch_namespace and self.parent_mod:
      if not layer.get_absolute_name().startswith("."):  # temp layer
        parent_mod, param_name = self.parent_mod
        mod_abs_name = naming.get_module_abs_id_name(parent_mod)
        torch_mod = naming.import_params_from_torch_namespace.get_module_by_abs_id_name(mod_abs_name)
        torch_param = getattr(torch_mod, param_name)
        values = torch_param.detach().numpy()
    if values is None:
      values = self.param.detach().numpy()
    assert isinstance(layer, VariableLayer)
    assert len(layer.params) == 1
    tf_param = list(layer.params.values())[0]
    session = tf.compat.v1.get_default_session()
    tf_param.load(values, session=session)
    return self.param


class Constant(Module):
  is_original_torch_module = False

  def __init__(self, value: Tensor):
    super(Constant, self).__init__()
    assert isinstance(value, Tensor)
    self.value = value

  def _get_batch_dim(self) -> Optional[int]:
    tensor = self.value
    batch_dims = [i for i, d in enumerate(tensor.shape) if d.is_batch_dim]
    assert len(batch_dims) <= 1  # not implemented otherwise
    if batch_dims:
      return batch_dims[0]
    return None

  def create_returnn_layer_dict(self, *inputs):  # ignore inputs
    tensor = self.value
    value = tensor.numpy()
    batch_axis = self._get_batch_dim()
    if batch_axis is not None:
      # We remove the batch axis now, and set with_batch_dim below.
      # But we need to check whether this is valid.
      value = numpy.moveaxis(value, batch_axis, 0)
      for i in range(1, value.shape[0]):
        numpy.testing.assert_equal(value[0], value[i])
      value = value[0]  # remove batch axis
    assert isinstance(value, numpy.ndarray)
    # Simplify representation in these simple cases.
    if not value.shape:  # scalar
      if value.dtype == "int32":
        value = int(value)
      elif value.dtype == "float32":
        value = float(value)
    d = {"class": "constant", "value": value}
    if batch_axis is not None:
      d["with_batch_dim"] = True
      naming = Naming.get_instance()
      call = naming.module_call_stack[-1]
      assert call.module.module is self
      # Add some of the network inputs as the dependency, to get the batch-dim.
      call.inputs_flat = [naming.inputs[0]]
    return d

  def make_output_tensor_from_returnn(self, inputs_flat: List[Tensor], layer: LayerBase) -> Tensor:
    naming = Naming.get_instance()
    tensor = self.value
    entry = naming.register_tensor(tensor)
    entry.returnn_data = layer.output
    batch_axis = self._get_batch_dim()
    if batch_axis is not None:
      # Batch axis will be moved to top.
      def _new_axis(old_axis: int) -> int:
        if old_axis == batch_axis:
          return 0
        if old_axis < batch_axis:
          return old_axis + 1
        return old_axis
      entry.returnn_axis_from_torch_axis = {i: _new_axis(i) for i in range(tensor.ndim)}
      entry.is_const = False  # this depends on the (dynamic) batch dim
    else:
      entry.returnn_axis_from_torch_axis = {i: i for i in range(tensor.ndim)}
    return tensor


__all__ = [
  key for (key, value) in sorted(globals().items())
  if not key.startswith("_")
  and getattr(value, "__module__", "") == __name__]





Verbosity = 0


_unique_prints = set()


def unique_print(txt: str):
  if Verbosity >= 100:  # always print, ignore unique
    print(txt)
    return
  if txt in _unique_prints:
    return
  _unique_prints.add(txt)
  print(txt)
